@@ -11,6 +11,34 @@ end
 luatexja.load_lua = load_lua
 function luatexja.load_module(name) require('ltj-' .. name.. '.lua') end
 
+do
+    local dnode = node.direct
+    local getfield, traverse = dnode.getfield, dnode.traverse
+    local node_new, set_attr, get_attr = dnode.new, dnode.set_attribute, dnode.get_attribute
+    local set_attrlist, get_attrlist = dnode.setattributelist, dnode.getattributelist
+    local unset_attr = dnode.unset_attribute
+    local attr_icflag = luatexbase.attributes['ltj@icflag']
+    local function node_inherit_attr(n, b, a)
+        if b or a then
+            local attrlist = get_attrlist(b or a)
+            local nic = get_attr(n, attr_icflag)
+            set_attrlist(n, attrlist); set_attr(n, attr_icflag, nic)
+            if b and a then
+                for na in traverse(attrlist) do
+                    local id = getfield(na, 'number')
+                    if id and id~=attr_icflag and getfield(na, 'value')~=get_attr(a, id) then
+                        unset_attr(n, id)
+                    end
+                end
+            end
+        end
+        return n
+    end
+    luatexja.node_inherit_attr = node_inherit_attr
+    luatexja.dnode_new = function (id, subtype, b, a)
+        return node_inherit_attr(node_new(id, subtype), b, a)
+    end
+end
 do
    local setfield = node.direct.setfield
    luatexja.setglue = node.direct.setglue or
@@ -295,13 +323,14 @@ do
    local to_node = node.direct.tonode
    local to_direct = node.direct.todirect
    local ensure_tex_attr = ltjb.ensure_tex_attr
-
+   local slide = node.slide
    -- mode = true iff main_process is called from pre_linebreak_filter
    local function main_process(head, mode, dir, gc)
       ensure_tex_attr(attr_icflag, 0)
       if gc == 'fin_row' then return head
       else
             start_time_measure('jfmglue')
+            slide(head);
             local p = ltjj.main(to_direct(head),mode, dir)
             stop_time_measure('jfmglue')
             return to_node(p)
@@ -332,11 +361,6 @@ do
    ltjb.add_to_callback('hpack_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
 end
 
--- cleanup process
-function luatexja.ext_cleanup()
-   ltjf.cleanup_size_cache(); ltjd.remove_end_whatsit()
-end
-
 -- lastnodechar
 do
    local id_glyph = node.id('glyph')
@@ -351,12 +375,12 @@ do
             r = tostring(n.char)
          end
       end
-      tex.sprint(r)
+      tex.sprint(-2, r)
    end
 end
 
 do
-    local cache_ver = 2 -- must be same as ltj-kinsoku.tex
+    local cache_ver = 3 -- must be same as ltj-kinsoku.tex
     local cache_outdate_fn = function (t) return t.version~=cache_ver end
     local t = ltjs.charprop_stack_table
     function luatexja.load_kinsoku()
@@ -526,7 +550,7 @@ local function debug_show_node_X(p,print_fn, limit, inner_depth)
             s = s .. ' stream=' .. p.stream
             print_fn(s)
             for i=1,#p.data do
-               print_fn(s .. '  [' .. i .. '] = ' .. tostring(p.data[i].csname))
+               print_fn(s .. '  [' .. i .. '] = ' .. tostring(p.data[i] and p.date[i].csname))
             end
          elseif p.subtype==16 then
             s = s .. ' mode=' .. p.mode .. ', literal="' .. p.data .. '"'
